<?php

namespace Drupal\entity_sync\Plugin\EntitySync\FieldTransformer;

use Drupal\entity_sync\Exception\InvalidConfigurationException;
use Drupal\entity_sync\FieldTransformer\PluginBase;
use Drupal\Core\Entity\ContentEntityInterface;

/**
 * Transformer that applies one or more callbacks to the value.
 *
 * Supported configuration properties:
 * - callbacks: (array, required) A numerical array containing one or more
 *   items, each defining a callback. Each callback is an associative array
 *   containing the following items:
 *   - callback (string, required): The callback to be called for processing the
 *     value being transformed, in one of the forms supported by
 *     `call_user_func_array()`.
 *   - arguments (string, optional, defaults to `value`): Defines the arguments
 *     that will be passed to the callback. Supported values are:
 *     - value: Only the value being transformed will be passed to the callback.
 *     - all: All variables available will be passed to the callback. See
 *       `$this::transformImportedValue()`,  `$this::transformExportedValue()`.
 *
 * If more than one callback is defined, they will be called one after the other
 * in the order provided. This is made possible as a convenience feature because
 * calling multiple time this plugin via a pipeline is syntactically cumbersome.
 *
 * The ability to pass all available variables as arguments to the callback is
 * provided for cases that writing a field transformer plugin is not
 * desirable. It is also the first step to deprecate direct callbacks in the
 * field mapping; they can be moved to a field transformer via this
 * plugin. However, these cases are discouraged and writing field transformers
 * should be the solution where possible. We therefore set the default setting
 * to providing only the value being transformed to the callback, since this
 * plugin is provided mostly for easily calling PHP functions e.g. strtolower
 * etc. for transforming the value. Many of such functions raise an error if
 * unexpected arguments are provided.
 *
 * @EntitySyncFieldTransformer(
 *   id = "callback"
 * )
 *
 * @link https://www.php.net/manual/en/function.call-user-func-array.php
 */
class Callback extends PluginBase {

  /**
   * {@inheritdoc}
   */
  protected function validateConfiguration() {
    $this->validateConfigurationRequiredProperties([
      'callbacks',
    ]);
  }

  /**
   * {@inheritdoc}
   */
  protected function transformImportedValue(
    $value,
    \stdClass $remote_entity,
    ?ContentEntityInterface $local_entity,
    array $field_info,
    array $context
  ) {
    return $this->transformValue(
      $value,
      [
        $value,
        $remote_entity,
        $local_entity,
        $field_info,
        $context,
      ]
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function transformExportedValue(
    $value,
    ContentEntityInterface $local_entity,
    $remote_entity_id,
    array $field_info,
    array $context
  ) {
    return $this->transformValue(
      $value,
      [
        $value,
        $local_entity,
        $remote_entity_id,
        $field_info,
        $context,
      ]
    );
  }

  /**
   * Returns the value as transformed by the callbacks.
   *
   * @param mixed $value
   *   The value to transform.
   * @param array $arguments
   *   The arguments to pass to the callbacks.
   *
   * @return mixed
   *   The transformed value.
   */
  protected function transformValue($value, array $arguments) {
    if ($value === NULL) {
      return NULL;
    }

    foreach ($this->configuration['callbacks'] as $callback) {
      if (($callback['arguments'] ?? 'value') === 'value') {
        $value = call_user_func_array($callback['callback'], [$value]);
      }
      elseif ($callback['arguments'] === 'all') {
        $value = call_user_func_array($callback, $arguments);
      }
      else {
        throw new InvalidConfigurationException(sprintf(
          'Unsupported option "%s" provided for callback arguments, expected "value" or "all".',
          $callback['arguments']
        ));
      }

      array_shift($arguments);
      array_unshift($arguments, $value);
    }

    return $value;
  }

}
